##ACTIONS

La configuration se fait en YAML dans le fichier `config/migration_action.yaml`

Une action s'execute avant ou après la migration générée par le mapping. Elle permet de faire des modifications en base de données, que ce soit sur la Source ou sur la Destination.

Un exemple d'utilisation est de vider les tables dans lesquelles on va insérer des données : on déclare dans `migration_action.yaml` qu'on va utiliser l'action `truncate_table_before_migration`  et on configure les tables à supprimer

###Créer une action

La création d'une action se fait en implémentant l'une des interfactions `PreMigrationActionInterface` ou `PostMigrationActionInterface` et en étendant la classe abstraite `AbstractMigrationAction`

```php
<?php

declare(strict_types=1);

namespace App\Service\MigrationAction;

class TruncateTableBeforeMigration extends AbstractMigrationAction implements PreMigrationActionInterface
{
    public function getType(): string
    {
            return 'truncate_table_before_migration';
    }

    public function execute(bool $commit): void
    {
        ...
    }
}
```

La valeur retournée par `getType` sera utilisée pour matcher la configuration configurée dans `migration_action.yaml`

Il est possible d'accéder à l'attribut `options` implémenté par `AbstractMigrationAction`, dans lequel les valeurs de configuration sont stockées.

La fonction execute récupère l'argument `$commit` qui détermine si les manipulations de base de données doivent être effectivement executée ou juste loggées

Exemple :

```php
    public function execute(bool $commit): void
    {
        foreach ($this->options['tables'] as $table) {                  // Récupération des tables configurées dans `migration_action.yaml`
            $sql = sprintf('DELETE FROM `%s`', $table);
            $this->logger->log(LogLevel::INFO, $sql);                   // Log de la query executée pour la visibilité dev

            if ($commit) {                                              // Si on est en mode execution, on déclenche la requête
                $this->connection->executeQuery($sql);
            }
        }
    }
```

Enfin, il est nécessaire d'ajouter la structure de la configuration de l'action dans la fonction `getConfigTreeBuilder` de `MigrationActionConfiguration` (voir [la documentation Symfony](https://symfony.com/doc/current/components/config/definition.html#array-nodes))

```php
class MigrationActionConfiguration implements ConfigurationInterface
{
    public function getConfigTreeBuilder()
    {
        $treeBuilder = new TreeBuilder('migration_mapping');

        $rootNode = $treeBuilder
            ->getRootNode();

        $rootNode
            ->children()
                ...
                ->arrayNode('truncate_table_before_migration')
                    ->scalarPrototype()->end()
                ->end()
                ...
            ->end()
        ;

        return $treeBuilder;
    }
}
```

###Actions disponibles

####TruncateTableBeforeMigration
**type**: `array liste`

Liste de table sur lesquelles executer un `DELETE FROM table WHERE 1` avant la migration. 

```yaml
dedi_sylius_migration:
  migration_actions:
    truncate_table_before_migration:
      - sylius_order
      - sylius_product
      - ...  
```

Lorsque la migration est executée en mode `group` (option `--group`) ou tables (argument `tables` sur la commande), le `truncate_table_before_migration` n'est pas executé.

Il faut alors utiliser l'option `truncate` sur le mapping de la table pour la vider.


####ExecuteQueryBeforeMigration
**type**: `array liste`

Liste de requête SQL à executer avant la migration

```yaml
dedi_sylius_migration:
  migration_actions:
    execute_query_before_migration:
      - 'ALTER TABLE sylius_locale ADD COLUMN legacy_id VARCHAR(255)'
      - 'UPDATE sylius_locale SET legacy_id = 1 WHERE id = 1'
      - ...
```

####CleanExtraLegacyColumnAfterMigration
**type**: `boolean`

Si activé (`true`), une fois la migration effectuée, le script va supprimer toutes les colonnes générées par la clé de config `extra.legacy_column_name` (voir [la documentation à ce sujet](MAPPING.md#migration_mapping--table_destinationextra))

```yaml
dedi_sylius_migration:
  migration_actions:
    clean_extra_legacy_column_after_migration: true
```
