<?php

declare(strict_types=1);

namespace Dedi\SyliusMigrationPlugin\Service\Sql;

use Dedi\SyliusMigrationPlugin\Config\MigrationActionConfig;
use Dedi\SyliusMigrationPlugin\Config\MigrationConfig;
use Dedi\SyliusMigrationPlugin\Config\MigrationMappingConfig;
use Dedi\SyliusMigrationPlugin\Dto\MigrationOption;
use Dedi\SyliusMigrationPlugin\Registry\HandlerRegistry;
use Dedi\SyliusMigrationPlugin\Service\Logger\MigrationLogger;
use Dedi\SyliusMigrationPlugin\Service\MigrationExecutorInterface;
use Dedi\SyliusMigrationPlugin\Service\Specification\IsDefinitionRunnable;
use Exception;
use Psr\Cache\CacheItemPoolInterface;
use Psr\Log\LogLevel;
use Symfony\Component\Console\Helper\ProgressBar;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Contracts\Cache\CacheInterface;

class MigrationExecutor implements MigrationExecutorInterface
{
    private readonly CacheInterface $cache;

    private readonly MigrationConfig $config;

    private readonly HandlerRegistry $handlerRegistry;

    private readonly MigrationActionConfig $migrationActionConfig;

    private readonly MigrationMappingConfig $mappingRegistry;

    private readonly ReaderInterface $reader;

    private readonly WriterInterface $writer;

    private readonly MigrationLogger $migrationLogger;

    private readonly IsDefinitionRunnable $isDefinitionRunnable;

    public function __construct(
        MigrationConfig $config,
        MigrationActionConfig $migrationActionConfig,
        MigrationMappingConfig $mappingRegistry,
        ReaderInterface $reader,
        WriterInterface $writer,
        MigrationLogger $migrationLogger,
        HandlerRegistry $handlerRegistry,
        IsDefinitionRunnable $isDefinitionRunnable,
        CacheInterface $cache,
    ) {
        $this->config = $config;
        $this->migrationActionConfig = $migrationActionConfig;
        $this->mappingRegistry = $mappingRegistry;
        $this->reader = $reader;
        $this->writer = $writer;
        $this->migrationLogger = $migrationLogger;
        $this->handlerRegistry = $handlerRegistry;
        $this->isDefinitionRunnable = $isDefinitionRunnable;
        $this->cache = $cache;
    }

    public function execute(MigrationOption $migrationOption, OutputInterface $output): void
    {
        $commit = !$migrationOption->isDryRun() && $this->config->isCommit();
        if (!$commit) {
            $this->migrationLogger->log(LogLevel::INFO, 'Running in dry mode');
            $this->defuse();
        }

        $this->migrationActionConfig->executePreMigrationActions($migrationOption);

        $numberOfTables = count($this->mappingRegistry->getTableDefinitions());
        foreach ($this->mappingRegistry->getTableDefinitions() as $index => $tableDefinition) {
            if (!$this->isDefinitionRunnable->isSatisfiedBy($tableDefinition, $migrationOption)) {
                continue;
            }

            $this->migrationLogger->info(
                $tableDefinition,
                sprintf(
                    'Import from %s to %s [%s/%s]',
                    '' === $tableDefinition->getSourceTableName() ? 'SQL Query' : $tableDefinition->getSourceTableName(),
                    $tableDefinition->getTargetTableName(),
                    $index + 1,
                    $numberOfTables,
                ),
            );

            if (!$this->writer->tableExists($tableDefinition, $tableDefinition->getTargetTableName())) {
                continue;
            }

            if (!$this->reader->tableExists($tableDefinition, $tableDefinition->getSourceTableName())) {
                continue;
            }

            if ($migrationOption->isTruncate() && null !== $tableDefinition->getTruncateMappingDefinition()) {
                $this->writer->truncate($tableDefinition);
            }

            $offset = $tableDefinition->getLimitMappingDefinition()->getOffset();
            $step = $tableDefinition->getLimitMappingDefinition()->getStep();
            $limit = $this->config->getDebugLimitStatements() > 0 ? $this->config->getDebugLimitStatements() : $tableDefinition->getLimitMappingDefinition()->getLimit();

            $this->reader->doCache($tableDefinition, $migrationOption, $offset, $limit, $step);
            $countAll = $this->reader->getResultCount($tableDefinition, $migrationOption, $limit);
            $count = 0;

            $this->writer->beforeMigrate($tableDefinition);

            $progressBar = new ProgressBar($output, $countAll);

            do {
                $arrData = $this->reader->read($tableDefinition, $migrationOption, $offset, $step);

                if (null !== $tableDefinition->getHandler()) {
                    $handler = $this->handlerRegistry->getHandler($tableDefinition->getHandler());
                    $errors = $handler->execute($arrData);

                    foreach ($errors as $error) {
                        $this->migrationLogger->error($tableDefinition, $error);
                    }
                    $dataCount = count($arrData);
                    $count += $dataCount;
                    $progressBar->advance($dataCount);
                } else {
                    foreach ($arrData as $data) {
                        try {
                            $this->writer->write($tableDefinition, $data);

                            if ($this->config->getDebugLimitStatements() > 0) {
                                ++$count;
                            }
                        } catch (Exception $exception) {
                            $this->migrationLogger->handleException($exception, $tableDefinition, $data);
                        }

                        if ($this->config->getDebugLimitStatements() <= 0) {
                            ++$count;
                        }

                        $progressBar->advance();

                        if ($count === $limit) {
                            break;
                        }
                    }
                }

                $offset += $step;
            } while (0 !== count($arrData) && $count !== $limit);

            $progressBar->finish();
            $output->writeln('');

            $this->writer->afterMigrate($tableDefinition);
            $this->reader->doClearCache($tableDefinition, $tableDefinition->getLimitMappingDefinition()->getOffset(), $limit, $step);
        }

        $this->migrationActionConfig->executePostMigrationActions($commit);
    }

    private function defuse()
    {
        $this->reader->defuse();
        $this->writer->defuse();
    }

    public function cleanUp(): void
    {
        if ($this->cache instanceof CacheItemPoolInterface) {
            $this->cache->clear();
        }
    }
}
