<?php

declare(strict_types=1);

namespace Dedi\SyliusMigrationPlugin\Service\Sql;

use Dedi\SyliusMigrationPlugin\Definition\TableMappingDefinition;
use Dedi\SyliusMigrationPlugin\Dto\DateFieldsMappingDefinition;
use Dedi\SyliusMigrationPlugin\Dto\MigrationOption;
use Dedi\SyliusMigrationPlugin\Dto\OrderMappingDefinition;
use Doctrine\DBAL\Query\QueryBuilder;
use LogicException;
use Symfony\Contracts\Cache\ItemInterface;

class Reader extends AbstractSqlProcessor implements ReaderInterface
{
    public function read(TableMappingDefinition $tableMappingDefinition, MigrationOption $migrationOption, int $offset, int $limit): array
    {
        return $this->cache->get($this->getCacheKey($tableMappingDefinition, $offset, $limit), function (ItemInterface $item) use ($tableMappingDefinition, $migrationOption, $offset, $limit): array {
            $this->logger->warning($tableMappingDefinition, 'Cache miss, fetching data from source');
            $item->expiresAfter(604800); // 1 week
            $sourceQueryBuilder = $this->resolveQueryBuilder($tableMappingDefinition);
            $sql = $this->resolveSqlStatement($tableMappingDefinition, $migrationOption, $sourceQueryBuilder, $offset, $limit);
            $parameters = $this->resolveParameters($tableMappingDefinition, $sourceQueryBuilder);
            $this->logger->debug($tableMappingDefinition, $sql);

            return $this->connection->executeQuery($sql, $parameters)->fetchAllAssociative();
        });
    }

    public function getResultCount(TableMappingDefinition $tableMappingDefinition, MigrationOption $migrationOption, int $limit): int
    {
        return $this->cache->get($this->getCountCacheKey($tableMappingDefinition), function (ItemInterface $item) use ($tableMappingDefinition, $migrationOption, $limit): int {
            $item->expiresAfter(604800); // 1 week
            $sourceQueryBuilder = $this->resolveCountQueryBuilder($tableMappingDefinition);
            $sql = sprintf('SELECT COUNT(*) FROM (%s) a', $this->resolveSqlStatement($tableMappingDefinition, $migrationOption, $sourceQueryBuilder, 0, $limit));
            $parameters = $this->resolveParameters($tableMappingDefinition, $sourceQueryBuilder);
            $this->logger->debug($tableMappingDefinition, $sql);

            $count = $this->connection->executeQuery($sql, $parameters)->fetchOne();

            return false !== $count ? (int) $count : 0;
        });
    }

    public function getLinesNoInTmpMappingTable(TableMappingDefinition $tableMappingDefinition, MigrationOption $migrationOption, array $excludeIds, int $limit): array
    {
        $sourceQueryBuilder = $this->resolveCountQueryBuilder($tableMappingDefinition);

        if (empty($excludeIds)) {
            $sql = sprintf('SELECT * FROM (%s) a', $this->resolveSqlStatement($tableMappingDefinition, $migrationOption, $sourceQueryBuilder, 0, $limit));
        } else {
            $sql = sprintf(
                'SELECT * FROM (%s) a WHERE a.%s NOT IN(%s)',
                $this->resolveSqlStatement($tableMappingDefinition, $migrationOption, $sourceQueryBuilder, 0, $limit),
                $tableMappingDefinition->getSourcePrimaryKey(),
                implode(',', $excludeIds),
            );
        }

        $parameters = $this->resolveParameters($tableMappingDefinition, $sourceQueryBuilder);
        $this->logger->debug($tableMappingDefinition, $sql);

        return $this->connection->executeQuery($sql, $parameters)->fetchAllAssociative();
    }

    private function addSelectPart(TableMappingDefinition $tableMappingDefinition, QueryBuilder $queryBuilder): void
    {
        foreach ($tableMappingDefinition->getColumnMappingDefinitions() as $columnMappingDefinition) {
            if (null !== $columnMappingDefinition->getSourceColumnName()) {
                $queryBuilder
                    ->addSelect($columnMappingDefinition->getSourceColumnName())
                ;
            }

            if (null !== $columnMappingDefinition->getExpr()) {
                if (null === $columnMappingDefinition->getAlias()) {
                    throw new LogicException('Expr configuration needs an Alias');
                }
                $queryBuilder
                    ->addSelect(sprintf('%s AS %s', $columnMappingDefinition->getExpr(), $columnMappingDefinition->getAlias()))
                ;
            }
        }
    }

    private function addSearchPart(TableMappingDefinition $tableMappingDefinition, QueryBuilder $queryBuilder): void
    {
        if ($tableMappingDefinition->hasFilterMapping()) {
            foreach ($tableMappingDefinition->getFilterMapping() as $filterMappingDefinition) {
                $pattern = $this->resolveSearchType($filterMappingDefinition->getType());

                $queryBuilder
                    ->andWhere(sprintf($pattern, $filterMappingDefinition->getFilterColumn(), $filterMappingDefinition->getFilterColumn()))
                    ->setParameter($filterMappingDefinition->getFilterColumn(), $filterMappingDefinition->getValue())
                ;
            }
        }
    }

    private function addOrderPart(TableMappingDefinition $tableMappingDefinition, QueryBuilder $queryBuilder): void
    {
        $orderMappingDefinition = $tableMappingDefinition->getOrderMappingDefinition();

        if ($orderMappingDefinition instanceof OrderMappingDefinition) {
            $queryBuilder->orderBy($orderMappingDefinition->getColumn(), $orderMappingDefinition->getSort());
        } else {
            $queryBuilder->orderBy($tableMappingDefinition->getSourcePrimaryKey(), 'DESC');
        }
    }

    private function createQueryBuilder(TableMappingDefinition $tableMappingDefinition): QueryBuilder
    {
        $sourceQueryBuilder = $this->connection->createQueryBuilder();
        $sourceQueryBuilder
            ->from($tableMappingDefinition->getSourceTableName())
            ->addSelect($tableMappingDefinition->getSourcePrimaryKey())
        ;

        return $sourceQueryBuilder;
    }

    private function resolveParameters(TableMappingDefinition $tableMappingDefinition, ?QueryBuilder $queryBuilder): array
    {
        if ('' !== $tableMappingDefinition->getSql()) {
            return [];
        }

        if (null !== $queryBuilder) {
            return $queryBuilder->getParameters();
        }

        throw new LogicException(sprintf('Could not resolve query parameters for table %s', $tableMappingDefinition->getTargetTableName()));
    }

    private function resolveQueryBuilder(TableMappingDefinition $tableMappingDefinition): ?QueryBuilder
    {
        if ('' !== $tableMappingDefinition->getSql()) {
            return null;
        }
        $sourceQueryBuilder = $this->createQueryBuilder($tableMappingDefinition);
        $this->addSelectPart($tableMappingDefinition, $sourceQueryBuilder);
        $this->addSearchPart($tableMappingDefinition, $sourceQueryBuilder);
        $this->addOrderPart($tableMappingDefinition, $sourceQueryBuilder);

        return $sourceQueryBuilder;
    }

    private function resolveCountQueryBuilder(TableMappingDefinition $tableMappingDefinition): ?QueryBuilder
    {
        if ('' !== $tableMappingDefinition->getSql()) {
            return null;
        }

        $sourceQueryBuilder = $this->createQueryBuilder($tableMappingDefinition);
        $this->addSearchPart($tableMappingDefinition, $sourceQueryBuilder);
        $this->addOrderPart($tableMappingDefinition, $sourceQueryBuilder);

        return $sourceQueryBuilder;
    }

    private function resolveSqlStatement(TableMappingDefinition $tableMappingDefinition, MigrationOption $migrationOption, ?QueryBuilder $queryBuilder, int $offset, int $limit): string
    {
        if ('' !== $tableMappingDefinition->getSql()) {
            $query = $this->addDatesPartToSql($tableMappingDefinition->getSql(), $tableMappingDefinition->getDateFieldsMappingDefinition(), $migrationOption);

            return $this->addLimitPartToSql($query, $offset, $limit);
        }

        if (null !== $queryBuilder) {
            $this->addLimitPartToQueryBuilder($queryBuilder, $offset, $limit);

            return $queryBuilder->getSQL();
        }

        throw new LogicException(sprintf('Could not resolve a runnable query for table %s', $tableMappingDefinition->getTargetTableName()));
    }

    private function resolveSearchType(string $type): string
    {
        switch ($type) {
            case 'comparison':
                return '%s = :%s';
            case 'not_comparison':
                return '%s != :%s';
            case 'like':
                return '%s LIKE :%s';
            case 'is_null':
                return '%s IS NULL';
            case 'in':
                return '%s IN (:%s)';
            default:
                throw new LogicException(sprintf('Unsupported type %s', $type));
        }
    }

    private function addLimitPartToSql(string $query, int $offset, int $limit): string
    {
        if ($limit === -1) {
            return $query;
        }

        return sprintf(trim($query) . ' LIMIT %s, %s', $offset, $limit);
    }

    private function addLimitPartToQueryBuilder(QueryBuilder $queryBuilder, $offset, $limit): void
    {
        if ($limit === -1) {
            return;
        }

        $queryBuilder
            ->setFirstResult($offset)
            ->setMaxResults($limit)
        ;
    }

    private function addDatesPartToSql(string $query, DateFieldsMappingDefinition $dateFieldsMappingDefinition, MigrationOption $migrationOption): string
    {
        $dateFields = $dateFieldsMappingDefinition->getFields();
        if (count($dateFields) === 0 || null === $migrationOption->getDateFrom()) {
            return $query;
        }

        $dateComparisons = [];
        foreach ($dateFields as $dateField) {
            $dateComparisons[] = sprintf('DATE(%s) >= "%s"', $dateField, $migrationOption->getDateFrom()->format('Y-m-d'));
        }

        $dateQuery = ' (' . implode(' OR ', $dateComparisons) . ') ';

        $havingPos = strripos($query, 'HAVING');
        if ($havingPos !== false) {
            $dateQuery .= ' AND ';

            return substr_replace($query, $dateQuery, $havingPos + 6, 0);
        }

        $dateQuery = ' HAVING ' . $dateQuery;

        $orderByPos = strripos($query, 'ORDER BY');
        if ($orderByPos !== false) {
            return substr_replace($query, $dateQuery, $orderByPos, 0);
        }

        return $query . $dateQuery;
    }

    public function doCache(TableMappingDefinition $tableMappingDefinition, MigrationOption $migrationOption, int $offset, int $limit, int $step): void
    {
        $this->logger->info($tableMappingDefinition, sprintf('Caching %s', $tableMappingDefinition->getTargetTableName()));
        $sourceQueryBuilder = $this->resolveQueryBuilder($tableMappingDefinition);

        $sql = $this->resolveSqlStatement($tableMappingDefinition, $migrationOption, $sourceQueryBuilder, 0, $limit);
        $parameters = $this->resolveParameters($tableMappingDefinition, $sourceQueryBuilder);
        $this->logger->debug($tableMappingDefinition, $sql);

        $arrData = $this->connection->executeQuery($sql, $parameters)->fetchAllAssociative();

        $this->cache->get($this->getCountCacheKey($tableMappingDefinition), function (ItemInterface $item) use ($arrData): int {
            $item->expiresAfter(604800); // 1 week

            return count($arrData);
        });

        $count = 0;
        do {
            /** @var array $data */
            $data = $this->cache->get($this->getCacheKey($tableMappingDefinition, $offset, $step), function (ItemInterface $item) use ($arrData, $offset, $step): array {
                $item->expiresAfter(604800); // 1 week

                return array_slice($arrData, $offset, $step);
            });
            $offset += $step;
            $count += count($data);
        } while (0 !== count($data) && $count !== $limit);
    }

    private function getCacheKey(TableMappingDefinition $tableMappingDefinition, int $offset, int $step): string
    {
        $str = '' != $tableMappingDefinition->getSql() ? $tableMappingDefinition->getSql() : $tableMappingDefinition->getTargetTableName();

        return hash('sha256', sprintf('%s_%s_%s', $str, $offset, $step));
    }

    private function getCountCacheKey(TableMappingDefinition $tableMappingDefinition): string
    {
        $str = '' != $tableMappingDefinition->getSql() ? $tableMappingDefinition->getSql() : $tableMappingDefinition->getTargetTableName();

        return hash('sha256', sprintf('count_%s', $str));
    }

    public function doClearCache(TableMappingDefinition $tableMappingDefinition, int $offset, int $limit, int $step): void
    {
        $this->logger->info($tableMappingDefinition, sprintf('Clearing cache %s', $tableMappingDefinition->getTargetTableName()));
        $this->cache->delete($this->getCountCacheKey($tableMappingDefinition));
        do {
            $data = $this->cache->get($this->getCacheKey($tableMappingDefinition, $offset, $step), function (ItemInterface $item): array {
                $item->expiresAfter(604800); // 1 week

                return [];
            });
            $result = $this->cache->delete($this->getCacheKey($tableMappingDefinition, $offset, $step));
            $offset += $step;
        } while (!empty($data) && $result);
    }
}
